/**
 * Copyright 2019 - MagabeLab (Tanzania). All Rights Reserved.
 * Author Edwin Magabe    edyma50@yahoo.com
 */

import 'package:petitparser/petitparser.dart';

import 'parsers.dart';
import 'simple_flatten_parser.dart';

///Parses attributes, if caseSensitive is set to false then case sensitivity is ignored,  for-example href will match HREF, hREF , Href etc
///And due to the characteristics of a parser returned by nonCaseSensitiveChars(), hrefXXX will also match HREF, hREF , Href etc
class AttributeParser extends Parser<String> {
  final String attributeKey;
  final hasValue;
  final bool caseSensitive;

  AttributeParser(
    this.attributeKey, {
    bool hasValue = true,
    bool caseSensitive = true,
  })  : hasValue = hasValue,
        caseSensitive = caseSensitive;

  @override
  Parser<String> copy() {
    return AttributeParser(attributeKey);
  }

  Parser<String> flatten([String message]) {
    return SimpleFlattenParser(this, message);
  }

  Parser _attributesWithValue() {
    final key =
        letter().plus().seq((letter() | digit() | pattern('-_~.')).star());
    final attrKey = key.seq(parsers.spaceOptional());
    final withValue = parsers
        .equal()
        .seq(parsers.spaceOptional())
        .seq(parsers.attributeValue());

    return attrKey.seq(withValue);
  }

  List<Map<String, String>> toAttributeMaps(List<String> attributes) {
    var attributeKeyAttributeValueMap = <String, String>{};
    var attributeKeyAttributeStringMap = <String, String>{};
    for (var attributeString in attributes) {
      attributeString = attributeString.trim();
      if (attributeString.isNotEmpty) {
        final index = attributeString.indexOf('=', 0);
        if (index >= 0) {
          final key = attributeString.substring(0, index).trim();
          final value =
              parsers.removeQuotes(attributeString.substring(index + 1));

          if (key.isNotEmpty) {
            //value can be null
            attributeKeyAttributeValueMap[key] = value;
            attributeKeyAttributeStringMap[key] = attributeString;
          }
        }
      }
    }
    return [attributeKeyAttributeValueMap, attributeKeyAttributeStringMap];
  }

  Parser matcher(String word) => parsers.nonCaseSensitiveChars(word);

  @override
  Result<String> parseOn(Context context) {
    final buffer = context.buffer;
    var position = context.position;

    if (position < buffer.length) {
      var end = buffer.indexOf('>', position);
      if (end >= 0) {
        var attributeContent =
            buffer.substring(position, end + 1); // 1 is the length of >

        if (attributeContent.isNotEmpty) {
          final tagStr = parsers.getParserResult(
              parser: parsers
                  .start()
                  .seq(parsers.spaceOptional())
                  .seq(letter().seq((letter() | digit()).star()))
                  .seq(parsers.spaceOptional()),
              input: attributeContent);

          if (tagStr.isNotEmpty) {
            // final tag = tagStr.replaceAll('<', '').trim();
            attributeContent = attributeContent.replaceAll(tagStr, '');
          }

          attributeContent = attributeContent.replaceAll('>', '');
          if (attributeContent.isNotEmpty) {
            final attributesWithValues = parsers.getParserResults(
                parser: _attributesWithValue(), input: attributeContent);

            var _tmp = attributeContent;
            attributesWithValues.forEach((attr) {
              _tmp = _tmp.replaceAll(attr, '');
            });
            if (_tmp.trim() == '/') {
              _tmp = '';
            } else {
              _tmp = _tmp.trim();
            }

            var attributesWithNoValuesList = _tmp.split('\s');

            final attributeMaps = toAttributeMaps(attributesWithValues);

            final attributeKeyAttributeValueMap = attributeMaps[0];

            final attributeKeyAttributeStringMap = attributeMaps[1];

            if (hasValue) {
              for (var _attributeKey in attributeKeyAttributeValueMap.keys) {
                var match = (attributeKey == null)
                    ? true
                    : (caseSensitive
                        ? (_attributeKey == attributeKey)
                        : matcher(_attributeKey).accept(attributeKey));

                final attributesString =
                    attributeKeyAttributeStringMap[_attributeKey];

                final index = buffer.indexOf(attributesString, position);
                if (index >= 0) {
                  //update position
                  position = index + attributesString.length;
                  if (match) {
                    return Success<String>(buffer, position, attributesString);
                  }
                }
              }
            } else {
              for (var _attributeKey in attributesWithNoValuesList) {
                var match = (attributeKey == null)
                    ? true
                    : (caseSensitive
                        ? (_attributeKey == attributeKey)
                        : matcher(_attributeKey).accept(attributeKey));

                final attributesString = _attributeKey;
                var index = buffer.indexOf(attributesString, position);
                if (index >= 0) {
                  //update position
                  position = index + attributesString.length;
                  if (match) {
                    return Success<String>(buffer, position, attributesString);
                  }
                }
              }
            }
          }
        }
      }

      return Failure<String>(
          context.buffer, position, 'Failed , No attribute found');
    } else {
      return Failure<String>(context.buffer, context.buffer.length,
          'Failed , reached the end of buffer');
    }
  }
}
